/*
 *	Simple Mail Queue Manager -- Back End
 *
 *	(c) 2000 Martin Mares <mj@ucw.cz>
 *
 *	Can be freely distributed and used under the terms of the GNU GPL.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <syslog.h>
#include <signal.h>
#include <time.h>
#include <getopt.h>
#include <dirent.h>
#include <grp.h>
#include <sys/stat.h>
#include <sys/wait.h>

int perscan_time = 300;			/* Once per 5 minutes force rescan of everything */
int tmp_timeout = 86400;		/* Delete temporary files older than one day automatically */
int debug;				/* Debugging mode */
int once;				/* Scan queue only once */

void
usage(void)
{
  fprintf(stderr, "Usage: usherd [-d] [-o] [-t <secs>] [-d <secs>]\n");
}

void
feed_child(char *cmd)
{
  setenv("HOME", HOMEDIR, 1);
  setenv("MAJORDOMO_CF", HOMEDIR "/majordomo.cf", 1);
  if (!cmd[0])
    execl("./majordomo", "./majordomo", NULL);
  else if (cmd[0] == '@')
    execl("./majordomo", "./majordomo", "-l", cmd+1, NULL);
  else
    {
      char *listname = alloca(strlen(cmd) + 8);
#if 0
      sprintf(listname, "%s-list", cmd);
#else
      sprintf(listname, "lists/%s", cmd);
#endif
      execl("./resend", "./resend", "-l", cmd, listname, NULL);
    }
  syslog(LOG_ERR, "exec: %m");
  exit(1);
}

int
handle_mail(char *name)
{
  int fd, i;
  pid_t pid;
  char buf[256], *c;

  if (debug)
    printf("Processing %s\n", name);

  close(0);
  fd = open(name, O_RDONLY);
  if (fd)
    {
      syslog(LOG_ERR, "Cannot open %s: %m", name);
      return 0;
    }
  i = read(fd, buf, sizeof(buf) - 1);
  if (i < 0)
    {
      syslog(LOG_ERR, "Cannot read from %s: %m", name);
      goto err;
    }
  buf[i] = 0;
  c = strchr(buf, '\n');
  if (!c)
    {
      syslog(LOG_ERR, "Malformed queue file %s", name);
      goto err;
    }
  *c++ = 0;
  if (debug)
    printf("Decoded destination <%s>\n", buf);
  lseek(fd, c - buf, SEEK_SET);

  pid = fork();
  if (pid < 0)
    {
      syslog(LOG_ERR, "fork: %m");
      goto err;
    }
  if (!pid)
    feed_child(buf);
  close(fd);
  while (waitpid(pid, &i, 0) < 0)
    ;
  if (WIFEXITED(i))
    {
      if (!WEXITSTATUS(i))
	return 1;
      syslog(LOG_ERR, "Child exited with rc=%d", WEXITSTATUS(i));
    }
  else if (WIFSIGNALED(i))
    syslog(LOG_ERR, "Child died on signal %d", WTERMSIG(i));	
  else
    syslog(LOG_ERR, "Child died of plague (symptom=%x)", i);
  return 0;

err:
  close(fd);
  return 0;
}

void
scan_queue(void)
{
  DIR *dir;
  struct dirent *e;

  if (debug)
    puts("Scanning queue...");
  dir = opendir("queue");
  while ((e = readdir(dir)) && !once)
    {
      char *a = alloca(8 + strlen(e->d_name));
      if (e->d_name[0] == '.')
	continue;
      sprintf(a, "queue/%s", e->d_name);
      if (!handle_mail(a))
	{
	  char *z = alloca(8 + strlen(e->d_name));
	  sprintf(z, "trash/%s", e->d_name);
	  if (rename(a, z) >= 0)
	    continue;
	  syslog(LOG_ERR, "Cannot rename %s to %s: %m", a, z);
	}
      if (unlink(a) < 0)
	syslog(LOG_ERR, "Error unlinking %s: %m", a);
    }
  closedir(dir);
}

void
scan_temps(void)
{
  DIR *dir;
  struct dirent *e;
  struct stat st;
  time_t now;

  if (!tmp_timeout)
    return;
  if (debug)
    puts("Scanning temporary files...");
  now = time(NULL);
  dir = opendir("tmp");
  while (e = readdir(dir))
    {
      char *a = alloca(8 + strlen(e->d_name));
      if (e->d_name[0] == '.')
	continue;
      sprintf(a, "tmp/%s", e->d_name);
      if (stat(a, &st) >= 0 && now - st.st_mtime > tmp_timeout)
	{
	  syslog(LOG_INFO, "Deleting obsolete temporary file %s", a);
	  if (unlink(a) < 0)
	    syslog(LOG_INFO, "Error deleting %s: %m", a);
	}
    }
  closedir(dir);
}

void
term_handler(int unused)
{
  once = 1;
}

void
sig_handler(int unused)
{
  scan_queue();
}

void
timer_handler(int unused)
{
  scan_queue();
  scan_temps();
  alarm(perscan_time);
}

int
main(int argc, char **argv)
{
  int c;
  struct sigaction sa;
  FILE *f;

  while ((c = getopt(argc, argv, "t:q:do")) >= 0)
    switch (c)
      {
      case 't':
	tmp_timeout = atol(optarg);
	break;
      case 'q':
	perscan_time = atol(optarg);
	break;
      case 'd':
	debug = 1;
	break;
      case 'o':
	once = 1;
	break;
      default:
	usage();
	return 1;
      }
  if (optind < argc)
    {
      usage();
      return 1;
    }

  if (chdir(HOMEDIR) < 0)
    {
      fprintf(stderr, "usherd: chdir(" HOMEDIR "): %m\n");
      return 1;
    }
  if (!debug && !once)
    {
      pid_t pid = fork();
      if (pid < 0)
	{
	  fprintf(stderr, "usherd: fork: %m\n");
	  return 1;
	}	
      if (pid)
	return 0;
      if (!getuid())
	{
	  setgid(MAJORDOMO_GID);
	  setgroups(0, NULL);
	  setuid(MAJORDOMO_UID);
	}
      setsid();
      setpgrp();
      close(1);
      open("/dev/null", O_RDWR);
    }

  openlog("usherd", LOG_PID, LOG_MAIL);

  bzero(&sa, sizeof(sa));
  sa.sa_handler = term_handler;
  sa.sa_flags = SA_RESTART;
  sigaction(SIGINT, &sa, NULL);
  sigaction(SIGTERM, &sa, NULL);
  sa.sa_handler = sig_handler;
  sa.sa_flags = SA_RESTART;
  sigaddset(&sa.sa_mask, SIGUSR1);
  sigaddset(&sa.sa_mask, SIGALRM);
  sigaction(SIGUSR1, &sa, NULL);
  sa.sa_handler = timer_handler;
  sigaction(SIGALRM, &sa, NULL);

  f = fopen("usherd.pid", "w");
  if (!f)
    {
      syslog(LOG_ERR, "Cannot create PID file: %m");
      return 1;
    }
  fprintf(f, "%d\n", getpid());
  fclose(f);

  kill(0, SIGALRM);		/* Force initial scanning and directory cleanup */
  while (!once)
    pause();
  unlink("usherd.pid");
  return 0;
}
